#!/usr/bin/env python3

import os
import json

class configInfo:

    def __init__(self, event_reporting_wake_interval, event_reporting_minimum_count, event_reporting_flush_interval, event_reporting_trap_exceptions, logging_verbose):
        self.__event_reporting_wake_interval = event_reporting_wake_interval
        self.__event_reporting_minimum_count = event_reporting_minimum_count
        self.__event_reporting_flush_interval = event_reporting_flush_interval
        self.__event_reporting_trap_exceptions = event_reporting_trap_exceptions
        self.__logging_verbose = logging_verbose

    def get_event_reporting_wake_interval(self):
        return self.__event_reporting_wake_interval
    event_reporting_wake_interval=property(get_event_reporting_wake_interval)

    def get_event_reporting_minimum_count(self):
        return self.__event_reporting_minimum_count
    event_reporting_minimum_count=property(get_event_reporting_minimum_count)

    def get_event_reporting_flush_interval(self):
        return self.__event_reporting_flush_interval
    event_reporting_flush_interval = property(get_event_reporting_flush_interval)

    def get_event_reporting_trap_exceptions(self):
        return self.__event_reporting_trap_exceptions
    event_reporting_trap_exceptions = property(get_event_reporting_trap_exceptions)
    
    def get_logging_verbose(self):
        return self.__logging_verbose
    logging_verbose = property(get_logging_verbose)

    def set(self):
        return (self.__event_reporting_wake_interval != None 
                and self.__event_reporting_minimum_count != None 
                and self.__event_reporting_flush_interval != None
                and self.__logging_verbose != None)

class ubiqConfiguration(configInfo):

    def load_config_file(self, config_file):
        try:
            with open(config_file) as json_file:
                config = json.load(json_file)
                self.__event_reporting_wake_interval = config['event_reporting']['wake_interval']
                self.__event_reporting_minimum_count = config['event_reporting']['minimum_count']
                self.__event_reporting_flush_interval = config['event_reporting']['flush_interval']
                self.__event_reporting_trap_exceptions = config['event_reporting']['trap_exceptions']
                self.__logging_verbose = config['logging']['verbose']
        except FileNotFoundError:
            # If file doesn't exist, use defaults
            self.set_defaults()

    def set_defaults(self):
        self.__event_reporting_wake_interval = 10
        self.__event_reporting_minimum_count = 50
        self.__event_reporting_flush_interval = 90
        self.__event_reporting_trap_exceptions = False
        self.__logging_verbose = False

    def __init__(self, config_file = None):

        self.__event_reporting_wake_interval = None
        self.__event_reporting_minimum_count = None
        self.__event_reporting_flush_interval = None
        self.__event_reporting_trap_exceptions = None
        self.__logging_verbose = None

        if (config_file == None):
            from os.path import expanduser
            home = expanduser("~")
            config_file = os.path.join(home, ".ubiq", "configuration")

        if os.path.exists(config_file):
            self.load_config_file(config_file)
        else:
            self.set_defaults()

        configInfo.__init__(self, self.__event_reporting_wake_interval , self.__event_reporting_minimum_count, self.__event_reporting_flush_interval, self.__event_reporting_trap_exceptions, self.__logging_verbose)



#/usr/bin/env python3

import base64
from .algo import ff1

from ..credentials import credentials

from .common import fmtInput, strConvertRadix, encKeyNumber, fmtOutput
from .common import fetchFFS, fetchKey, fetchCurrentKeys

class Encryption:
    def __del__(self):
        # todo: overwrite 'unwrapped_data_key'
        return

    def __init__(self, creds, ffs):
        if not creds.set():
            raise RuntimeError("credentials not set")
        
        self._creds = creds

        # If the host does not begin with either http or https
        # insert https://
        self._host = creds.host
        if (not self._host.lower().startswith('http')):
            self._host = "https://" + self._host
        self._host += '/api/v0/'

        self._papi = creds.access_key_id
        self._sapi = creds.secret_signing_key
        self._srsa = creds.secret_crypto_access_key

        self._ffs = fetchFFS(self._host, self._papi, self._sapi, ffs)
        self._key = fetchKey(self._host,
                             self._papi, self._sapi,
                             self._srsa,
                             ffs)

        if self._ffs['encryption_algorithm'] == 'FF1':
            self._algo = ff1.Context(
                self._key['unwrapped_data_key'],
                base64.b64decode(self._ffs['tweak']),
                self._ffs['tweak_min_len'], self._ffs['tweak_max_len'],
                len(self._ffs['input_character_set']),
                self._ffs['input_character_set'])
        else:
            raise RuntimeError('unsupported algorithm: ' +
                               self._ffs['encryption_algorithm'])

    def Cipher(self, pt, twk = None):
        pth = self._ffs['passthrough']
        ics = self._ffs['input_character_set']
        ocs = self._ffs['output_character_set']

        fmt, pt = fmtInput(pt, pth, ics, ocs)

        ct = self._algo.Encrypt(pt, twk)

        ct = strConvertRadix(ct, ics, ocs)
        ct = encKeyNumber(ct, ocs,
                          self._key['key_number'],
                          self._ffs['msb_encoding_bits'])
        
        self._creds.add_event(dataset_name=self._ffs['name'], dataset_group_name="", billing_action="encrypt",
                dataset_type="structured", key_number=self._key['key_number'], count=1)
        
        return fmtOutput(fmt, ct, pth)
    
    def CipherForSearch(self, pt, twk=None):
        keys = fetchCurrentKeys(self._host,
                            self._papi, self._sapi,
                            self._srsa,
                            self._ffs['name'])
        
        pth = self._ffs['passthrough']
        ics = self._ffs['input_character_set']
        ocs = self._ffs['output_character_set']

        fmt, pt = fmtInput(pt, pth, ics, ocs)


        searchCipher = []
        for _, (key_num, key) in enumerate(keys.items()):
            algo = ff1.Context(
                key['unwrapped_data_key'],
                base64.b64decode(self._ffs['tweak']),
                self._ffs['tweak_min_len'], self._ffs['tweak_max_len'],
                len(ics),
                ics)
            ct = algo.Encrypt(pt, twk)
            ct = strConvertRadix(ct, ics, ocs)
            ct = encKeyNumber(ct, ocs,
                          key_num,
                          self._ffs['msb_encoding_bits'])
            searchCipher.append(fmtOutput(fmt, ct, pth))

        return searchCipher



def Encrypt(creds, ffs, pt, twk = None):
    return Encryption(creds, ffs).Cipher(pt, twk)

def EncryptForSearch(creds, ffs, pt, twk = None):
    return Encryption(creds, ffs).CipherForSearch(pt, twk)



#!/usr/bin/env python3

import base64
import cryptography.hazmat.backends
import cryptography.hazmat.primitives.hashes
import cryptography.hazmat.primitives.hmac
import email.utils
import http.client
import requests.auth
import time
import urllib.parse

from .version import VERSION

class http_auth(requests.auth.AuthBase):
    """HTTP Authentication for the Ubiq Platform

    This module implements HTTP authentication for the Ubiq platform
    via message signing as described by the IETF httpbis-message-signatures
    draft specification.

    The module is intended to be used in conjunction with the `requests`
    library as in `requests.get(url, auth=http_auth(id, key))`.
    """

    def __init__(self, access_id, access_key):
        """
        access_id:
            a string containing the client's public API key (pAPI)
        access_key:
            a string containing the client's secret API key (sAPI)
        """

        self.access_id = access_id
        self.access_key = access_key

    def __call__(self, r):
        """
        The `requests` library calls this function prior to sending
        the HTTP request out over the network. This function calculates
        the signature for the message, adding the Signature header
        to contain the data. Certain HTTP headers are required for
        signature calculation and will be added by this code as
        necessary. The modified request is returned.
        """

        r.headers['User-Agent'] = 'ubiq-python/' + VERSION

        # the '(request-target)' is part of the signed data.
        # it's value is 'http_method path?query'

        parsed = urllib.parse.urlparse(r.url)
        r.headers['Content-type'] = 'application/json'
        req_tgt = r.method.lower() + ' ' + parsed.path
        if parsed.query:
            req_tgt += '?' + parsed.query

        # the time at which the signature was created
        # expressed as the unix epoch

        created = str(int(time.time()))

        # the requests library doesn't typically add the Host
        # header. it needs to be present to be part of the
        # signature. the port value is not included if it is
        # the default port for the scheme

        if not r.headers.get('Host'):
            r.headers['Host'] = parsed.hostname
            if parsed.port:
                if ((parsed.scheme == 'http' and
                     parsed.port != http.client.HTTP_PORT) or
                    (parsed.scheme == 'https' and
                     parsed.port != http.client.HTTPS_PORT)):
                    r.headers['Host'] += ':' + str(parsed.port)

        # the Date field is required for the signature

        if not r.headers.get('Date'):
            r.headers['Date'] = email.utils.formatdate(
                timeval=None, localtime=False, usegmt=True)

        # the Digest header is always included/overridden by
        # this code. it is a hash of the body of the http message
        # and is always present even if the body is empty

        hash_sha512 = cryptography.hazmat.primitives.hashes.Hash(
            cryptography.hazmat.primitives.hashes.SHA512(),
            backend=cryptography.hazmat.backends.default_backend())

        if r.body:
            hash_sha512.update(r.body)

        r.headers['Digest']  = 'SHA-512='
        r.headers['Digest'] += base64.b64encode(
            hash_sha512.finalize()).decode('utf-8')

        #
        # sign the message. the signature is an hmac of the
        # headers listed below
        #

        hmac_sha512 = cryptography.hazmat.primitives.hmac.HMAC(
            self.access_key.encode('utf-8'),
            cryptography.hazmat.primitives.hashes.SHA512(),
            backend=cryptography.hazmat.backends.default_backend())

        headers = []

        # the (request-target) and (created) are faux headers defined
        # by the message signature spec. they are added to the list
        # of "real" headers to make the code below simpler and then
        # removed from the request after the hmac has been updated

        r.headers['(request-target)'] = req_tgt
        r.headers['(created)'] = created

        # include the specified headers in the hmac calculation. each
        # header is of the form 'header_name: header value\n'
        #
        # included headers are also added to an ordered list of headers
        # which is included in the message

        for name in ['(created)',
                     '(request-target)',
                     'Content-Length',
                     'Content-Type',
                     'Date',
                     'Digest',
                     'Host']:
            if r.headers.get(name):
                headers.append(name.lower())
                hmac_sha512.update(
                    (name.lower() + ': ' + r.headers[name] + '\n').
                    encode('utf-8'))

        del r.headers['(created)']
        del r.headers['(request-target)']

        # build the Signature header itself

        r.headers['Signature']  = 'keyId="' + self.access_id + '"'
        r.headers['Signature'] += ', algorithm="hmac-sha512"'
        r.headers['Signature'] += ', created=' + created
        r.headers['Signature'] += ', headers="' + ' '.join(headers) + '"'
        r.headers['Signature'] += ', signature="'
        r.headers['Signature'] += base64.b64encode(
            hmac_sha512.finalize()).decode('utf-8')
        r.headers['Signature'] += '"'

        return r



#!/usr/bin/env python3

import base64
import http
import json
import requests
import struct
import urllib.error

import cryptography.exceptions as crypto_exceptions
import cryptography.hazmat.primitives as crypto
from cryptography.hazmat.backends import default_backend as crypto_backend

from . import UBIQ_HOST
from .auth import http_auth
from .algorithm import algorithm

class decryption:
    def _endpoint_base(self):
        return self._host + '/api/v0'

    def reset(self):
        """Reset the internal state of the decryption object

        This function can be called at any time to abort an existing
        decryption operation. It is also called by internal functions
        when a new decryption requires a different key than the one
        used by the previous decryption.
        """
        if hasattr(self, '_key'):
            if self._key['uses'] > 0:
                requests.patch(
                    self._endpoint_base() +
                    '/decryption/key/' + self._key['finger_print'] + '/' + self._key['session'],
                    data=json.dumps(
                        { 'uses': self._key['uses'] }).encode('utf-8'),
                    auth=http_auth(self._papi, self._sapi))

            del self._key

    def __del__(self):
        try:
            self.reset()
        except:
            pass

    def __init__(self, creds):
        """Initialize the encryption object

        papi:
            The client's public API key (used to identify the
            client to the server)
        sapi:
            The client's secret API key (used to authenticate HTTP requests)
        srsa:
            The client's secret RSA encryption key/password (used to decrypt
            the client's RSA key from the server). This key is not retained
            by this object.
        host:
            A string of the form 'host[:port]' with the []'s denoting an
            optional portion of the string indicating the server to which
            to make the request.
        """

        if not creds.set():
            raise RuntimeError("credentials not set")

        # If the host does not begin with either http or https
        # insert https://
        self._host = creds.host
        if (not self._host.lower().startswith('http')):
            self._host = "https://" + self._host

        self._papi = creds.access_key_id
        self._sapi = creds.secret_signing_key
        self._srsa = creds.secret_crypto_access_key

    def begin(self):
        """Begin the decryption process

        returns:
            any plain text produced by the call
        """

        # this interface does not take any cipher text in its arguments
        # in an attempt to maintain an API that corresponds to the
        # encryption object. in doing so, the work that can take place
        # in this function is limited. without any data, there is no
        # way to determine which key is in use or decrypt any data.
        #
        # this function simply throws an error if starting an decryption
        # while one is already in progress, and initializes the internal
        # buffer, otherwise

        if hasattr(self, '_key') and 'dec' in self._key:
            raise RuntimeError("decryption already in progress")

        self._buf = b''
        return b''

    def update(self, data):
        """Decrypt cipher text

        Cipher text must be passed to this function in the order in
        which it was output from the encryption.update function.

        data:
            (A portion of) the cipher text to be decrypted.  data
            value has to be contained in a bytes, bytearray or memoryview object.

        returns:
            any plain text produced by the call
        """

        if not isinstance(data, (bytes, bytearray, memoryview)):
            raise RuntimeError("Data must be bytes, bytearray, or memoryview objects")

        #
        # each encryption has a header on it that identifies the algorithm
        # used  and an encryption of the data key that was used to encrypt
        # the original plain text. there is no guarantee how much of that
        # data will be passed to this function or how many times this
        # function will be called to process all of the data. to that end,
        # this function buffers data internally, when it is unable to
        # process it.
        #
        # the function buffers data internally until the entire header is
        # received. once the header has been received, the encrypted data
        # key is sent to the server for decryption. after the header has
        # been successfully handled, this function always decrypts all of
        # the data in its internal buffer *except* for however many bytes
        # are specified by the algorithm's tag size. see the end() function
        # for details.
        #

        self._buf += data
        pt = b''

        # if there is no key or 'dec' member of key, then the code
        # is still trying to build a complete header

        if not hasattr(self, '_key') or not 'dec' in self._key:
            fmt = '!BBBBH'
            fmtlen = struct.calcsize(fmt)

            # does the buffer contain enough of the header to
            # determine the lengths of the initialization vector
            # and the key?

            if len(self._buf) >= fmtlen:
                ver, flags, alg, veclen, keylen = struct.unpack(
                    fmt, self._buf[:fmtlen])

                # For VER 0, lsb of indicates AAD or not
                if (ver != 0) or (flags & ~algorithm.UBIQ_HEADER_V0_FLAG_AAD):
                    raise RuntimeError('invalid encryption header')

                # does the buffer contain the entire header?

                if len(self._buf) >= fmtlen + veclen + keylen:

                    # Get the Header for AAD purposes.  Only needed if
                    # version != 0, but get it now anyways
                    aad = self._buf[:fmtlen + veclen + keylen]
                    # extract the initialization vector and the key
                    vec = self._buf[fmtlen:fmtlen + veclen]
                    key = self._buf[fmtlen + veclen:fmtlen + veclen + keylen]

                    # remove the header from the buffer
                    self._buf = self._buf[fmtlen + veclen + keylen:]

                    # generate a local identifier for the key
                    sha = crypto.hashes.Hash(
                        crypto.hashes.SHA256(), backend=crypto_backend())
                    sha.update(key)
                    client_id = sha.finalize()

                    # if the object already has a key (from a previous
                    # decryption), is the key in this header the same as
                    # that previous one?
                    #
                    # if not, clear out the existing key
                    if hasattr(self, '_key'):
                        if self._key['client_id'] != client_id:
                            self.reset()

                    # if the object (still) has a key, then it can be
                    # reused--see below. if not, then request a decryption
                    # of the key in the current header from the server
                    if not hasattr(self, '_key'):
                        url = self._endpoint_base() + '/decryption/key'
                        response = requests.post(
                            url,
                            data=json.dumps(
                                {
                                    'encrypted_data_key': base64.b64encode(
                                        key).decode('utf-8')
                                }).encode('utf-8'),
                            auth=http_auth(self._papi, self._sapi))
                        if response.status_code != http.HTTPStatus.OK:
                            raise urllib.error.HTTPError(
                                url, response.status_code,
                                http.HTTPStatus(response.status_code).phrase,
                                response.headers, response.content)

                        content = json.loads(response.content.decode('utf-8'))

                        self._key = {}

                        self._key['algo']      = algorithm(alg)
                        # the client's id for recognizing key reuse
                        self._key['client_id'] = client_id
                        # the server's id for sending updates
                        self._key['finger_print'] = content['key_fingerprint']
                        self._key['session'] = content['encryption_session']

                        # decrypt the client's private key (sent
                        # by the server)
                        prvkey = crypto.serialization.load_pem_private_key(
                            content['encrypted_private_key'].encode('utf-8'),
                            self._srsa.encode('utf-8'),
                            crypto_backend())
                        # use the private key to decrypt the data key
                        self._key['raw'] = prvkey.decrypt(
                            base64.b64decode(content['wrapped_data_key']),
                            crypto.asymmetric.padding.OAEP(
                                mgf=crypto.asymmetric.padding.MGF1(
                                    algorithm=crypto.hashes.SHA1()),
                                algorithm=crypto.hashes.SHA1(),
                                label=None))

                        # this key hasn't been used (yet)
                        self._key['uses'] = 0

                    # if the key object exists, create a new decryptor
                    # with the initialization vector from the header and
                    # the decrypted key (which is either new from the
                    # server or cached from the previous decryption). in
                    # either case, increment the key usage
                    if hasattr(self, '_key'):
                        self._key['dec'] = self._key['algo'].decryptor(
                            self._key['raw'], vec)
                        self._key['uses'] += 1

                        if (flags & algorithm.UBIQ_HEADER_V0_FLAG_AAD):
                            self._key['dec'].authenticate_additional_data(aad)

        # if the object has a key and a decryptor, then decrypt whatever
        # data is in the buffer, less any data that needs to be saved to
        # serve as the tag.
        if hasattr(self, '_key') and 'dec' in self._key:
            sz = len(self._buf) - self._key['algo'].len['tag']
            if sz > 0:
                pt = self._key['dec'].update(self._buf[:sz])
                self._buf = self._buf[sz:]

        return pt

    def end(self):
        """Finish a decryption

        returns:
            any plain text produced by the call
        """
        # the update function always maintains tag-size bytes in
        # the buffer because this function provides no data parameter.
        # by the time the caller calls this function, all data must
        # have already been input to the decryption object.

        pt = b''

        try:
            # determine how much of the buffer contains data
            # and how much of it contains the tag
            sz = len(self._buf) - self._key['algo'].len['tag']

            if sz < 0:
                # there's not enough data in the buffer for a complete tag
                raise crypto_exceptions.InvalidTag
            elif sz == 0:
                # the buffer contains exactly the right amount of data
                # for a complete tag. if the tag length is zero, just
                # finalize the decryption
                if self._key['algo'].len['tag'] == 0:
                    pt = self._key['dec'].finalize()
                else:
                    pt = self._key['dec'].finalize_with_tag(self._buf)
            else:
                # this is a logic error that can't occur based
                # on the logic in the update function. the update
                # function never leaves more data than the tag
                # size in the buffer once the header has been
                # successfully parsed.
                raise AssertionError
        finally:
            del self._key['dec']
            del self._buf

        return pt

def decrypt(creds, data):
    """Simple decryption interface
    papi:
        The client's public API key (used to identify the
        client to the server)
    sapi:
        The client's secret API key (used to authenticate HTTP requests)
    srsa:
        The client's secret RSA encryption key/password (used to decrypt
        the client's RSA key from the server).
    data:
        A byte string containing the cipher text to be decrypted
    host:
        A string of the form 'host[:port]' with the []'s denoting an
        optional portion of the string indicating the server to which
        to make the request.

    returns:
        the entire cipher text that can be passed to the decrypt function
    """
    dec = decryption(creds)
    creds.add_event(dataset_name="", dataset_group_name="", billing_action="decrypt",
                dataset_type="unstructured", key_number=0, count=1)
    return dec.begin() + dec.update(data) + dec.end()



#!/usr/bin/env python3

import base64
import http
import json
import requests
import urllib

from ..auth import http_auth
from ubiq_security_fpe import ffx


import cryptography.hazmat.primitives as crypto
from cryptography.hazmat.backends import default_backend as crypto_backend

def strConvertRadix(s, ics, ocs):
    return ffx.NumberToString(len(ocs), ocs,
                              ffx.StringToNumber(len(ics), ics, s),
                              len(s))

def fmtInput(s, pth, ics, ocs):
    fmt = ''
    trm = ''
    for c in s:
        if c in pth:
            fmt += c
        else:
            fmt += ocs[0]
            if c in ics:
                trm += c
            else:
                raise RuntimeError('invalid input character')
    return fmt, trm

def encKeyNumber(s, ocs, n, sft):    
    return ocs[ocs.find(s[0]) + (int(n) << sft)] + s[1:]

def decKeyNumber(s, ocs, sft):
    charBuf = s[0]
    encoded_value = ocs.find(charBuf)
    key_num = encoded_value >> sft

    return ocs[encoded_value - (key_num << sft)] + s[1:], key_num

def fmtOutput(fmt, s, pth):
    o = ''
    for c in fmt:
        if c not in pth:
            o, s = o + s[0], s[1:]
        else:
            o += c

    if len(s) > 0:
        raise RuntimeError('mismatched format and output strings')

    return o

def fetchFFS(host, papi, sapi, ffs):
    if (not papi in fetchFFS.cache or
        not ffs in fetchFFS.cache[papi]):
        url = host + 'ffs'
        url += '?ffs_name=' + ffs
        url += '&papi=' + papi
        resp = requests.get(url, auth=http_auth(papi, sapi))
        if resp.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(
                url, resp.status_code,
                http.HTTPStatus(resp.status_code).phrase,
                resp.headers, resp.content)
        if not papi in fetchFFS.cache:
            fetchFFS.cache[papi] = {}
        fetchFFS.cache[papi][ffs] = json.loads(resp.content.decode())

    return fetchFFS.cache[papi][ffs]
fetchFFS.cache = {}

def flushFFS(papi = None, ffs = None):
    if papi == None:
        fetchFFS.cache = {}
    elif papi in fetchFFS.cache:
        if ffs == None:
            del fetchFFS.cache[papi]
        elif ffs in fetchFFS.cache[papi]:
            del fetchFFS.cache[papi][ffs]

def fetchKey(host, papi, sapi, srsa, ffs, n = -1):
    if (not papi in fetchKey.cache or
        not ffs in fetchKey.cache[papi] or
        not n in fetchKey.cache[papi][ffs]):
        url = host + 'fpe/key'
        url += '?ffs_name=' + ffs
        url += '&papi=' + papi
        if n >= 0:
            url += '&key_number=' + str(n)
        resp = requests.get(url, auth=http_auth(papi, sapi))
        if resp.status_code != http.HTTPStatus.OK:
            raise urllib.error.HTTPError(
                url, resp.status_code,
                http.HTTPStatus(resp.status_code).phrase,
                resp.headers, resp.content)
        key = json.loads(resp.content.decode())

        prvkey = crypto.serialization.load_pem_private_key(
            key['encrypted_private_key'].encode(), srsa.encode(),
            crypto_backend())

        key['unwrapped_data_key'] = prvkey.decrypt(
            base64.b64decode(key['wrapped_data_key']),
            crypto.asymmetric.padding.OAEP(
                mgf=crypto.asymmetric.padding.MGF1(
                    algorithm=crypto.hashes.SHA1()),
                algorithm=crypto.hashes.SHA1(),
                label=None))

        if not papi in fetchKey.cache:
            fetchKey.cache[papi] = {}
        if not ffs in fetchKey.cache[papi]:
            fetchKey.cache[papi][ffs] = {}

        # the -1 entry points to the "current" key at the
        # server. it is cached so that the next caller that
        # wants the "current" key can get it, but it should
        # be timed-out occasionally in case the "current"
        # pointer changes at the server.
        #
        # that timeout is future work
        if n == -1:
            # -1 can be an index because keys are stored
            # in a dictionary, not a list
            fetchKey.cache[papi][ffs][n] = key

        # also cache the key at its "real" identifier
        n = int(key['key_number'])
        fetchKey.cache[papi][ffs][n] = key

    return fetchKey.cache[papi][ffs][n]
fetchKey.cache = {}

def allKeysToNInCache(papi, ffs, n):
    present = True
    for i in range(0,n+1):
        present = present and (i in fetchKey.cache[papi][ffs])
    return present

def fetchAllKeys(host, papi, sapi, srsa, ffs):
    url=f"{host}fpe/def_keys?ffs_name={ffs}&papi={papi}"
    resp = requests.get(url, auth=http_auth(papi, sapi))

    if resp.status_code != http.HTTPStatus.OK:
        raise urllib.error.HTTPError(
            url, resp.status_code,
            http.HTTPStatus(resp.status_code).phrase,
            resp.headers, resp.content)
    keys = json.loads(resp.content.decode())

    if not papi in fetchKey.cache:
        fetchKey.cache[papi] = {}
    if not ffs in fetchKey.cache[papi]:
        fetchKey.cache[papi][ffs] = {}

    prvkey = crypto.serialization.load_pem_private_key(
        keys[ffs]['encrypted_private_key'].encode(), srsa.encode(),
        crypto_backend())
    
    for i, enc_key in enumerate(keys[ffs]['keys']):
        if i in fetchKey.cache[papi][ffs]:
            continue

        key = {
            'encrypted_private_key': keys[ffs]['encrypted_private_key'],
            'wrapped_data_key': enc_key
        }
        key['unwrapped_data_key'] = prvkey.decrypt(
            base64.b64decode(key['wrapped_data_key']),
            crypto.asymmetric.padding.OAEP(
                mgf=crypto.asymmetric.padding.MGF1(
                    algorithm=crypto.hashes.SHA1()),
                algorithm=crypto.hashes.SHA1(),
                label=None))
        fetchKey.cache[papi][ffs][i] = key

def fetchCurrentKeys(host, papi, sapi, srsa, ffs,):
    fetchAllKeys(host, papi, sapi, srsa, ffs)

    return {key_num: key for key_num, key in sorted(fetchKey.cache[papi][ffs].items()) if key_num not in [-1]}

def flushKey(papi = None, ffs = None, n = None):
    if papi == None:
        fetchKey.cache = {}
    elif papi in fetchKey.cache:
        if ffs == None:
            del fetchKey.cache[papi]
        elif ffs in fetchKey.cache[papi]:
            if n == None:
                del fetchKey.cache[papi][ffs]
            elif n in fetchKey.cache[papi][ffs]:
                del fetchKey.cache[papi][ffs][n]



#/usr/bin/env python3

import base64

from ..credentials import credentials


from .common import fmtInput, strConvertRadix, decKeyNumber, fmtOutput
from .common import fetchFFS, fetchKey
from ubiq_security_fpe import ff1

class Decryption:
    def __del__(self):
        # todo: overwrite 'unwrapped_data_key'
        return

    def __init__(self, creds, ffs):
        if not creds.set():
            raise RuntimeError("credentials not set")
        
        self._creds = creds

        # If the host does not begin with either http or https
        # insert https://
        self._host = creds.host
        if (not self._host.lower().startswith('http')):
            self._host = "https://" + self._host
        self._host += '/api/v0/'

        self._papi = creds.access_key_id
        self._sapi = creds.secret_signing_key
        self._srsa = creds.secret_crypto_access_key

        self._ffs = fetchFFS(self._host, self._papi, self._sapi, ffs)

    def Cipher(self, ct, twk = None):
        pth = self._ffs['passthrough']
        ics = self._ffs['input_character_set']
        ocs = self._ffs['output_character_set']

        fmt, ct = fmtInput(ct, pth, ocs, ics)
        ct, n = decKeyNumber(ct, ocs, self._ffs['msb_encoding_bits'])
        if not hasattr(self, '_key') or self._key['key_number'] != n:
            self._key = fetchKey(self._host,
                                 self._papi, self._sapi, self._srsa,
                                 self._ffs['name'], n)
            if self._ffs['encryption_algorithm'] == 'FF1':
                self._ctx = ff1.Context(
                    self._key['unwrapped_data_key'],
                    base64.b64decode(self._ffs['tweak']),
                    self._ffs['tweak_min_len'], self._ffs['tweak_max_len'],
                    len(ics), ics)
            else:
                raise RuntimeError('unsupported algorithm: ' +
                                   self._ffs['encryption_algorithm'])
        ct = strConvertRadix(ct, ocs, ics)

        pt = self._ctx.Decrypt(ct, twk)

        self._creds.add_event(dataset_name=self._ffs['name'], dataset_group_name="", billing_action="decrypt",
            dataset_type="structured", key_number=n, count=1)
        return fmtOutput(fmt, pt, pth)

def Decrypt(creds, ffs, ct, twk = None):
    return Decryption(creds, ffs).Cipher(ct, twk)



#!/usr/bin/env python3

UBIQ_HOST = 'api.ubiqsecurity.com'

from .auth import http_auth
from .encrypt import encryption, encrypt
from .decrypt import decryption, decrypt
from .credentials import credentials, configCredentials
